# -*- coding: utf-8 -*-
# Copyright 2022 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/license.html.
# A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/authors.html.
"""Tools for calculating time-dependent manybody Green functions."""

import copy
import numpy as np
import collections.abc

import kwantspectrum
from . import manybody, _logging, _common


__all__ = ['GreenFunction']


# set module logger
logger = _logging.make_logger(name=__name__)
log_func = _logging.log_func(logger)


class _GreenGeneric():

    # generic green function of the form
    # g(t0, t1) = i \sum_\alpha \int \frac{dE}{2 \pi} f_\alpha(E) \psi_{\alpha, E}(t0) \psi^*_{\alpha, E}(t1)
    # if f is the fermi function we get the lesser green function g^<,
    # whereas for f=1 we get (the negative) retarded function g^R

    def __init__(self, *args, refine=True, **kwargs):

        # no refine on individual wave function based on density
        self._state0 = manybody.State(*args, **kwargs, refine=False)
        self._state1 = manybody.State(*args, **kwargs, refine=False)

        self.comm = self._state0.comm

        if refine:
            self.refine_intervals()

        # by default, return the result of the higher-order rule, which in
        # our convention has to be the last element of the weight array
        self.return_element = -1

    def evolve(self, time0, time1):
        '''Evolve the Greenfunction foreward in time up to G(time0, time1)'''

        self._state0.evolve(time0)
        self._state1.evolve(time1)

    def evaluate(self, i, j, root=0):
        '''Evaluate the Greenfunction at the site positions G_{i,j}'''

        result = self._calc_green_element(i, j, root=root)

        if self.return_element is None:
            return result
        try:
            return result[self.return_element]
        except Exception:  # catch `None` result on all non-MPI root ranks
            return result

    def refine_intervals(self, atol=1E-5, rtol=1E-5, limit=2000,
                         sites=None, intervals=None):
        '''Refine the manybody integral of all diagonal G_{i,i} elements
        Parameters
        ----------
        sites : sequence of tuples (i, j) with i, j type integer, optional
            By default, refine is done on all diagonal G_{i,i} elements
            If present, refine will be done for G_{i,j} elements where i, j
            are elements of the tuples (i, j). This replaces the
            ```error_op``` argument in ```tkwant.manybody.State.refine()```
        other parameters as for ```tkwant.manybody.State.refine()```
        '''

        assert _common.is_type(atol, 'real_number')
        if atol < 0:
            raise ValueError('atol={} is negative.'.format(atol))
        assert _common.is_type(rtol, 'real_number')
        if rtol < 0:
            raise ValueError('rtol={} is negative.'.format(rtol))
        assert _common.is_type(limit, 'integer')
        if limit <= 1:
            raise ValueError('limit={} must be > 1.'.format(limit))
        tol = min(1E-14, 0.5 * (atol + rtol))

        time0 = self._state0.time
        time1 = self._state1.time

        if sites is None:
            sites = [(i, i) for i in range(len(self._state0.extended_syst.syst.sites))]

        def green_with_error(_intervals):
            green = []
            errors = []
            for interval in _intervals:
                green_ = []
                errors_ = []
                for elem in sites:
                    error, kronrod = self._error_estimate_quadpack(interval,
                                                                   elem[0], elem[1],
                                                                   return_estimate=True,
                                                                   tol=0.5 * (atol + rtol))
                    green_.append(kronrod)
                    errors_.append(error)
                green.append(green_)
                errors.append(errors_)
            return np.array(green), np.array(errors)

        if intervals is None:
            intervals = self._state0.get_intervals()
        else:
            intervals = copy.deepcopy(intervals)
            if not isinstance(intervals, collections.abc.Iterable):
                intervals = [intervals]

        # refine only intervals with Gauss-Kronrod quadrature
        intervals[:] = [interval for interval in intervals
                        if interval.quadrature == 'kronrod']

        results, errors = green_with_error(intervals)

        result = np.sum(results, axis=0)  # sum of the integrals over the subintervals
        errsum = np.sum(errors, axis=0)  # sum of the errors over the subintervals

        i = 0
        while True:  # loop until converged

            errbnd = np.maximum(atol, rtol * np.abs(result))  # requested accuracy

            # order all intervals by error, decreasing order
            try:
                max_error_per_interval = np.max(errors - errbnd[np.newaxis, :], axis=1)
            except:
                max_error_per_interval = errors - errbnd

            error_idx = np.argsort(max_error_per_interval)[::-1]

            intervals = [intervals[i] for i in error_idx]
            results = results[error_idx]
            errors = errors[error_idx]
            max_error_per_interval = max_error_per_interval[error_idx]

            errmax = errors[0]
            resultmax = results[0]

            logger.info('refinement step={}, time0={}, time1={}, max errsum={}, '
                        'min errbnd={}, total number of intervals={}'.
                        format(i, time0, time1, np.max(errsum), np.min(errbnd), len(intervals)))
            for j, (inte, err) in enumerate(zip(intervals, errors)):
                logger.debug("interval_num={}, {}, max error={}".format(j, inte, np.max(err)))

            if (errsum <= errbnd).all():  # converged
                logger.info('refinement converged')
                break

            if len(intervals) >= limit:
                logger.warning('maximum number of intervals reached')
                break

            # bisect the interval with the largest error
            interval_largest_error = intervals.pop(0)
            new_intervals = manybody._split_interval(interval_largest_error, 2)
            intervals += new_intervals

            # update the intervals stored in the solver, evolve to current time
            self._state0._remove_interval(interval_largest_error)
            self._state1._remove_interval(interval_largest_error)
            for interval in new_intervals:
                self._state0._add_interval(interval, tol=tol)
                self._state1._add_interval(interval, tol=tol)
            self.evolve(time0, time1)

            # recalculate the error and observable estimate
            new_results, new_errors = green_with_error(new_intervals)

            results = np.append(results, new_results, axis=0)
            errors = np.append(errors, new_errors, axis=0)

            results = np.delete(results, 0, axis=0)
            errors = np.delete(errors, 0, axis=0)

            result = result + new_results[0] + new_results[1] - resultmax
            errsum = errsum + new_errors[0] + new_errors[1] - errmax
            i += 1

        return np.max(errsum), intervals, errors

    def estimate_error(self, i, j, intervals=None, full_output=False):
        '''Estimate the error of the manybody integral'''

        sum_up = False
        if intervals is None:
            intervals = self._state0.get_intervals()
            sum_up = True
        if isinstance(intervals, collections.abc.Iterable):
            errors_per_interval = np.array([self._error_estimate_quadpack(interval, i, j)
                                            for interval in intervals])
            if sum_up:
                errors = np.sum(errors_per_interval, axis=0)
                if not full_output:
                    errors = np.max(errors)
            else:
                if full_output:
                    errors = errors_per_interval
                else:
                    errors = np.array([np.max(err) for err in errors_per_interval])
        else:
            if full_output:
                errors = estimate(intervals, error_op)
            else:
                errors = np.max(estimate(intervals, error_op))
        return errors

    def _calc_green_element(self, i, j, keys=None, root=0, return_integrand=False):
        '''Evaluation of a Green function element and the integrand'''

        # the mpi rank distribution is deterministic, so the we will have
        # the same keys on the same mpi rank

        local_keys = self._state0.manybody_wavefunction.psi.local_keys()
        if keys is None:
            keys = local_keys
        else:  # intersection between given keys and keys on MPI rank
            keys = list(set(keys) & set(local_keys))

        integral = 0  # local integral part on mpi process

        # calculate weighted sum (partly on every rank)
        if return_integrand:
            integrand = {}
            for key in keys:
                psi_0 = self._state0.manybody_wavefunction.psi.local_data()[key].psi()
                psi_1 = np.conj(self._state1.manybody_wavefunction.psi.local_data()[key].psi())
                psi2 = psi_0[i] * psi_1[j]
                weight = 1j * self._state0.manybody_wavefunction.tasks[key].weight
                phys_weight = 1j * self._state0.manybody_wavefunction.tasks[key].phys_weight
                integral += np.outer(weight, psi2)
                integrand[key] = phys_weight * psi2
        else:
            for key in keys:
                psi_0 = self._state0.manybody_wavefunction.psi.local_data()[key].psi()
                psi_1 = np.conj(self._state1.manybody_wavefunction.psi.local_data()[key].psi())
                psi2 = psi_0[i] * psi_1[j]
                weight = 1j * self._state0.manybody_wavefunction.tasks[key].weight
                integral += np.outer(weight, psi2)

        if root is None:
            result = self.comm.allreduce(integral)
        else:
            result = self.comm.reduce(integral, root=root)
        result = np.squeeze(result)

        if return_integrand:
            return result, integrand
        return result

    def _evaluate_interval(self, interval, i, j, root=None, return_integrand=False):
        """Evaluate the statistical average over the given interval.

        Parameters
        ----------
        interval : `tkwant.manybody.Interval`
            Interval over which the manybody sum should be calculated.
        i, j : integer
            Green function element
        root : int or None, optional
            root receive the result, other rank receive None.
            If root is None all ranks receive the result.
        return_integrand : bool, optional
            If true, return also the integrand of the integral.

        Returns
        -------
        result : `~numpy.ndarray`
            The green function element, integrated over the interval.
        integrand : dict, optional
            The integrand to the integral. Only returned if ``return_integrand``
            it true.
        """
        keys = self._state0._get_keys_from_interval(interval)
        return self._calc_green_element(i, j, keys=keys, root=root,
                                        return_integrand=return_integrand)

    def _error_estimate_quadpack(self, interval, i, j, return_estimate=False, tol=1e-5):
        '''Estimate an manybody integral error of a given interval for the G_{i,j} element.'''

        if not interval.quadrature == 'kronrod':
            raise ValueError('quadpack error estimate works only on '
                             'Gauss-Kronrod quadrature intervals')

        (gauss, kronrod), func = self._evaluate_interval(interval, i, j,
                                                         return_integrand=True)

        dk = interval.kmax - interval.kmin
        assert dk > 0
        assert tol > 0
        kronrod_scaled = kronrod / dk

        integral = 0
        for key, f in func.items():
            kronrod_weight = self._state0.manybody_wavefunction.tasks[key].math_weight[1]
            integral += kronrod_weight * np.abs(f - kronrod_scaled)
        ik = self.comm.allreduce(integral)

        ik = np.where(ik > 0, ik, tol)  # ik can become exactly zero

        assert (ik > 0).all()
        tmp = 200 * np.abs(gauss - kronrod) / ik
        error = np.squeeze(ik * np.minimum(1, tmp * np.sqrt(tmp)))
        if return_estimate:
            return error, kronrod
        return error


class GreenFunction():

    # This class has a lazy init, evolve and refine_intervals routine.
    # The calls to these methods is saved in a joblist and the actual calls
    # are only executed before the the evaluate() method of the stored green
    # functions (green_lesser, green_retarded) is called.
    # This is for performance reasons and guarantees that only the function which
    # is really required (green_lesser and/or green_retarded) is actually calculated.

    '''

    Parameters
    ----------
    init positional and keyword arguments as ```tkwant.manybody.State```.
    The remaining parameters are:
    i, j : integer
        Green function element G_{i,j}
    time0, time1 : float, float
        Green function times G_(time0, time1)
    root : int or None, optional
        MPI root receive the result, other rank receive None.
        If root is None all ranks receive the result.
    There is no ```evaluate()``` method, but a method to evaluate
    each Green function type, such as ```.lesser()``` for G^<, ```.greater()```
    for G^> and so on.
    '''

    def __init__(self, *args, **kwargs):

        logger.debug('init GreenFunction')

        self._args = args
        self._kwargs = kwargs
        self._jobs_lesser = []
        self._jobs_retarded = []
        self._green_lesser = None
        self._green_retarded = None
        self._time0 = 0
        self._time1 = 0

    def evolve(self, time0, time1):
        if time0 < self._time0:
            raise ValueError("time0={} < previous time0={}".format(time0, self._time0))
        if time1 < self._time1:
            raise ValueError("time1={} < previous time1={}".format(time1, self._time1))
        self._jobs_lesser.append(('_evolve_lesser', (time0, time1), {}))
        self._jobs_retarded.append(('_evolve_retarded', (time0, time1), {}))
        self._time0 = time0
        self._time1 = time1

    def refine_intervals(self, *args, **kwargs):
        self._jobs_lesser.append(('_refine_lesser', args, kwargs))
        self._jobs_retarded.append(('_refine_retarded', args, kwargs))

    def lesser(self, i, j, root=0):
        if self._green_lesser is None:
            self._init_lesser()
        self._do_jobs(self._jobs_lesser)
        return self._green_lesser.evaluate(i, j, root)

    def retarded(self, i, j, root=0):
        if self._time0 < self._time1:
            return 0
        if self._green_retarded is None:
            self._init_retarded()
        self._do_jobs(self._jobs_retarded)
        res = self._green_retarded.evaluate(i, j, root)
        if res == None:
            return
        return - res

    def advanced(self, i, j, root=0):
        if self._time0 >= self._time1:
            return 0
        if self._green_retarded is None:
            self._init_retarded()
        self._do_jobs(self._jobs_retarded)
        return self._green_retarded.evaluate(j, i, root)

    def greater(self, i, j, root=0):
        # G^R − G^A = G^> − G^<
        lesser = self.lesser(i, j, root)
        if self._time0 >= self._time1:
            gra = self.retarded(i, j, root)
            sgn = 1
        else:
            gra = self.advanced(i, j, root)
            sgn = -1
        if lesser == None:
            return
        return lesser + sgn * gra

    def ordered(self, i, j, root=0):
        if self._time0 >= self._time1:
            return self.greater(i, j, root)
        return self.lesser(i, j, root)

    def anti_ordered(self, i, j, root=0):
        if self._time0 >= self._time1:
            return self.lesser(i, j, root)
        return self.greater(i, j, root)

    def keldysh(self, i, j, root=0):
        greater = self.greater(i, j, root)
        lesser = self.lesser(i, j, root)
        if greater == None:
            return
        return greater + lesser

    def _init_lesser(self):
        logger.debug('init G^<')
        self._green_lesser = _GreenGeneric(*self._args, **self._kwargs)

    def _init_retarded(self):

        logger.debug('init G^R')

        #  get the occuption form the arguments (positional or keyword)
        if len(self._args) > 2:
            occupations = self._args[2]
        else:
            occupations = self._kwargs.get('occupations', None)

        #  fill up all bands up to the highest energy
        # respect that not all bands/leads might be taken into acount
        if occupations is None:
            occupations = manybody.lead_occupation(chemical_potential=None)
        else:
            try:  # sequence of occupations
                full_occupations = []
                for occ in occupations:
                    if occ is None:  # if a lead is not counted, it should stay unoccupied
                        full_occupations.append(None)
                    else:
                        full_occupations.append(manybody.lead_occupation(chemical_potential=None))
                occupations = full_occupations
            except Exception:  # if occupations is not iterable
                occupations = manybody.lead_occupation(chemical_potential=None)

        #  change the occuption in args (positional) or kwargs (keyword)
        kwargs = copy.deepcopy(self._kwargs)  # dict is mutable !
        if len(self._args) > 2:
            arglist = list(self._args)
            arglist[2] = occupations
            args = tuple(arglist)
        else:
            args = self._args
            kwargs['occupations'] = occupations

        self._green_retarded = _GreenGeneric(*args, **kwargs)

    def _do_jobs(self, joblist):
        # loop over all jobs to do, reinitialize list afterwards
        for jobname, args, kwargs in joblist:
            logger.debug('jobname={}, args={}, kwargs={}'.format(jobname, args, kwargs))
            job = getattr(self, jobname)
            job(*args, **kwargs)
        joblist.clear()

    def _refine_lesser(self, *args, **kwargs):
        logger.debug('refine G^<')
        self._green_lesser.refine_intervals(*args, **kwargs)

    def _refine_retarded(self, *args, **kwargs):
        logger.debug('refine G^R')
        self._green_retarded.refine_intervals(*args, **kwargs)

    def _evolve_lesser(self, time0, time1, **kwargs):
        logger.debug('evolve G^<')
        self._green_lesser.evolve(time0, time1)

    def _evolve_retarded(self, time0, time1, **kwargs):
        logger.debug('evolve G^R')
        self._green_retarded.evolve(time0, time1)

    def estimate_error_lesser(self, i, j, intervals=None, full_output=False):
        self._do_jobs(self._jobs_lesser)
        return self._green_lesser.estimate_error(i, j, intervals, full_output)



# -*- coding: utf-8 -*-
# Copyright 2016-2018 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/license.html.
# A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/authors.html.

import pytest
from math import cos, sin
import itertools as it
import functools as ft
import numpy as np
from numpy.testing import assert_array_almost_equal
import tinyarray as ta
import scipy
import scipy.integrate._ode as ode
import scipy.sparse.linalg as spla
import kwant
import pickle
import cmath
from scipy.special import erf
import collections
import copy


from ... import leads
from .. import solvers, kernels, WaveFunction, ScatteringStates


solvers = {
    'scipy': (
        solvers.Scipy,
        [
            dict(integrator=ode.dopri5),
            dict(integrator=ode.dop853),
        ]
    ),
}


# TODO: focus on default kernel only for params
available_kernels = {
    'scipy': kernels.Scipy
}


# available_kernels = {
#    'scipy': kernels.Scipy,
#    'simple': kernels.Simple,
#    'sparseblas': getattr(kernels, 'SparseBlas', None)
# }


def get_solver_kernel(solver_t, kernel_t):
    skip = False
    reasons = []
    if kernel_t in ('sparseblas',):
        if not hasattr(kernels, 'SparseBlas'):
            skip = True
            reasons.append('No sparse BLAS available.')
    skip = pytest.mark.skipif(skip, reason=', '.join(reasons))

    kernel_t = available_kernels[kernel_t]
    solver_t, solver_params = solvers[solver_t]

    return [pytest.param(ft.partial(solver_t, **p), kernel_t, marks=skip)
            for p in solver_params]


def all_solvers_kernels():
    return pytest.mark.parametrize(
        ('solver_type', 'kernel_type'),
        list(it.chain.from_iterable(
            it.starmap(get_solver_kernel,
                       it.product(solvers.keys(), available_kernels.keys())))))


def all_solvers():  # test only with the pure python kernel
    return pytest.mark.parametrize(
        ('solver_type', 'kernel_type'),
        list(it.chain.from_iterable(
            it.starmap(get_solver_kernel,
                       it.product(solvers.keys(), ['scipy'])))))


def make_simple_lead(lat, N):
    I0 = ta.identity(lat.norbs)
    syst = kwant.Builder(kwant.TranslationalSymmetry((-1, 0)))
    syst[(lat(0, j) for j in range(N))] = 4 * I0
    syst[lat.neighbors()] = -1 * I0
    return syst


def make_complex_lead(lat, N):
    I0 = ta.identity(lat.norbs)
    syst = kwant.Builder(kwant.TranslationalSymmetry((-1, 0)))
    syst[(lat(0, j) for j in range(N))] = 4 * I0
    syst[kwant.HoppingKind((0, 1), lat)] = -1 * I0
    syst[(lat(0, 0), lat(1, 0))] = -1j * I0
    syst[(lat(0, 1), lat(1, 0))] = -1 * I0
    syst[(lat(0, 2), lat(1, 2))] = (-1 + 1j) * I0
    return syst


class RandomOnsite:
    def __init__(self, I0):
        self.I0 = I0

    def __call__(self, site, time, salt):
        return (4 + kwant.digest.uniform(site.tag, salt=salt)) * self.I0


def make_system(lat, N):
    I0 = ta.identity(lat.norbs)
    syst = kwant.Builder()
    syst[(lat(i, j) for i in range(N) for j in range(N))] = RandomOnsite(I0)
    syst[lat.neighbors()] = -1 * I0
    return syst


def make_td_system(lat, N, td_onsite):
    I0 = ta.identity(2)
    syst = kwant.Builder()
    square = it.product(range(N), range(N))
    syst[(lat(i, j) for i, j in square)] = td_onsite
    syst[lat.neighbors()] = -1 * I0
    return syst


def make_system_with_leads(lat, N, lead_maker):
    syst = make_system(lat, N)
    syst.attach_lead(lead_maker(lat, N))
    syst.attach_lead(lead_maker(lat, N).reversed())
    return syst


def make_td_system_with_leads(lat, N, lead_maker, td_onsite):
    syst = make_td_system(lat, N, td_onsite)
    syst.attach_lead(lead_maker(lat, N))
    syst.attach_lead(lead_maker(lat, N).reversed())
    return syst


@pytest.mark.integtest  # run only as integration test
@all_solvers_kernels()
def test_finite_time_independent(solver_type, kernel_type):
    N = 10
    salt = '1'
    lat = kwant.lattice.square(norbs=2)
    syst = make_system(lat, N)
    fsyst = syst.finalized()

    ### Start from eigenstate
    # no time dependence and starting from an eigenstate
    # φ means that we should evolve to φ * exp(-1j * E * t)
    H0 = fsyst.hamiltonian_submatrix(params={'time': 0, 'salt': salt})
    evals, evecs = spla.eigsh(H0)
    psi_st = evecs[:, 0]
    E = evals[0]
    # Solve "trivially" using (H0-E) @ psibar + W(t) @ psibar.
    # The actual ODE that we solve is 0=0, as W(t) = 0 for this system.
    psi = WaveFunction.from_kwant(syst=fsyst, params={'salt': salt}, psi_init=psi_st, energy=E,
                                  solver_type=solver_type, kernel_type=kernel_type)
    # Test that we get the same result when we use (H0 + W(t)) * psi directly.
    # We can do this because the system is *finite*
    psi_alt = WaveFunction.from_kwant(syst=fsyst, psi_init=psi_st, params={'salt': salt},
                                      solver_type=solver_type, kernel_type=kernel_type)
    # Construct a wavefunction directly from Hamiltonian matrices with sparse H0
    H0_sparse = fsyst.hamiltonian_submatrix(params={'time': 0, 'salt': salt}, sparse=True)
    W = kernels.PerturbationExtractor(fsyst, time_name='time', time_start=0,
                                      params=dict(salt=salt))
    psi_sparse = WaveFunction(H0=H0_sparse, W=W, psi_init=psi_st, energy=E, params={'salt': salt},
                              solver_type=solver_type, kernel_type=kernel_type)

    for time in (0., 1., 5.):
        psi.evolve(time)
        psi_alt.evolve(time)
        psi_sparse.evolve(time)
        assert np.allclose(psi.psi(), psi_st * np.exp(-1j * E * time))
        assert np.allclose(psi_alt.psi(), psi.psi())
        assert np.allclose(psi_sparse.psi(), psi_st * np.exp(-1j * E * time))


# TODO: pickle works only with the python kernel for the moment.
# Make it working for all kernels such that the test runs with
# @all_solvers_kernels().
@all_solvers()
def test_finite_time_independent_restart_pickle(solver_type, kernel_type):
    N = 10
    salt = '1'
    lat = kwant.lattice.square(norbs=2)
    syst = make_system(lat, N)
    fsyst = syst.finalized()

    ### Start from eigenstate
    # no time dependence and starting from an eigenstate
    # φ means that we should evolve to φ * exp(-1j * E * t)
    H0 = fsyst.hamiltonian_submatrix(params={'time': 0, 'salt': salt})
    evals, evecs = spla.eigsh(H0)

    ### test restarting a calculation
    psi_rand = (np.random.random((H0.shape[0],)) +
                1j * np.random.random((H0.shape[0],)))
    for psi_init, E in [(psi_rand, None), (evecs[:, 0], evals[0])]:
        psi = WaveFunction.from_kwant(syst=fsyst, psi_init=psi_init, params={'salt': salt}, energy=E,
                                      solver_type=solver_type, kernel_type=kernel_type)
        psi.evolve(2)
        saved = pickle.dumps(psi)
        psi2 = pickle.loads(saved)
        psi.evolve(5)
        psi2.evolve(5)
        assert np.allclose(psi.psi(), psi2.psi())


@pytest.mark.integtest  # run only as integration test
@all_solvers_kernels()
def test_finite_time_dependent(solver_type, kernel_type):
    N = 10
    salt = '1'
    lat = kwant.lattice.square(norbs=2)
    I0 = ta.identity(2)
    SX = ta.array([[0, 1], [1, 0]])
    uniform = kwant.digest.uniform

    def td_onsite(site, time, salt):
        static_part = (4 + uniform(site.tag, salt=salt)) * I0
        td_part = uniform(site.tag, salt=salt + '1') * SX * cos(time)
        return static_part + td_part

    syst = make_td_system(lat, N, td_onsite)
    fsyst = syst.finalized()
    nsites, _, norbs = fsyst.site_ranges[-1]
    sx_op = kwant.operator.Density(fsyst, np.array([[0, 1], [1, 0]]))

    H0 = fsyst.hamiltonian_submatrix(params={'time': 0, 'salt': salt})
    evals, evecs = spla.eigsh(H0)
    psi_st = evecs[:, 0]
    E = evals[0]

    perturb_types = [kernels.PerturbationExtractor, kernels.PerturbationInterpolator]

    for perturb in perturb_types:
        ### system has s_z symmetry: test for conservation of s_z
        psi = WaveFunction.from_kwant(syst=fsyst, psi_init=psi_st, energy=E, params={'salt': salt},
                                      perturbation_type=perturb,
                                      solver_type=solver_type, kernel_type=kernel_type)
        density = sx_op(psi.psi())
        should_be = np.sum(density)
        for time in (1., 5., 10.):
            density = sx_op(psi.psi())
            assert np.isclose(
                should_be,
                np.sum(density)
            )


@pytest.mark.integtest  # run only as integration test
@all_solvers_kernels()
def test_infinite_time_independent(solver_type, kernel_type):
    N = 10
    tmax = 5
    salt = '1'
    E = 1.
    lat = kwant.lattice.square(norbs=2)
    syst = make_system_with_leads(lat, N, make_simple_lead)
    fsyst = syst.finalized()
    norbs = fsyst.site_ranges[-1][-1]

    # naively, one would expect that num_buffer_cells = tmax * vmax / 2
    # where the factor 2 accounts for the forth and back travel of the pulse
    # in the lead. due to the non-trivial vacuum in terms of the fermi see,
    # the pulse "feels" the the border of the lead boundary, even before.
    # this is why we estimate num_buffer_cells much larger (here by a factor four of above length),
    # in order to have small reflections with simple boundaries.
    vmax = 2 * np.linalg.norm(fsyst.leads[0].inter_cell_hopping(), ord=2)
    num_buffer_cells = 2 * int(vmax * tmax)

    boundaries = [leads.SimpleBoundary(num_buffer_cells)] * 2

    ### Starting from the trivial state
    # solution should be zero everywhere
    zeros = np.zeros((norbs,), complex)
    psi = WaveFunction.from_kwant(fsyst, zeros, params={'salt': salt}, boundaries=boundaries,
                                  solver_type=solver_type, kernel_type=kernel_type)
    for time in (0., tmax / 4, tmax / 2, tmax):
        psi.evolve(time)
        assert np.allclose(psi.psi(), np.zeros((norbs,)))

    ### Start from eigenstate
    # No time dependence and starting from an eigenstate
    # φ means that we should evolve to φ * exp(-1j * E * t).
    # The actual ODE that we solve is 0=0, as W(t) = 0 for this system.
    # Call to Kwant requires time arg to be passed explicitly
    psi_st = kwant.wave_function(fsyst, energy=E, params={'time': 0, 'salt': salt})(0)[0]
    psi = WaveFunction.from_kwant(fsyst, psi_st, params={'salt': salt},
                                  boundaries=boundaries, energy=E,
                                  solver_type=solver_type, kernel_type=kernel_type)
    for time in (0., tmax / 4, tmax / 2, tmax):
        psi.evolve(time)
        assert np.allclose(psi.psi(), psi_st * np.exp(-1j * E * time))

    norbs = fsyst.site_ranges[-1][-1]
    psi_random = np.random.random((norbs,)) + 1j * np.random.random((norbs,))

    ### test that boundary conditions are OK by testing that solutions
    ### match when the boundary conditions are extended
    extended_boundaries = [leads.SimpleBoundary(2 * num_buffer_cells)] * 2
    psi = WaveFunction.from_kwant(fsyst, psi_random, boundaries=boundaries, params={'salt': salt},
                                  solver_type=solver_type, kernel_type=kernel_type)
    psi_ext = WaveFunction.from_kwant(fsyst, psi_random, boundaries=extended_boundaries, params={'salt': salt},
                                      solver_type=solver_type, kernel_type=kernel_type)
    size = psi.psibar.shape[0]  # largest subset of psi_ext and psi
    for time in (0., tmax / 4, tmax / 2, tmax):
        psi.evolve(time)
        psi_ext.evolve(time)
        assert np.allclose(psi_ext.psi()[:size], psi.psi()[:size])


# TODO: pickle works only with the python kernel for the moment.
# Make it working for all kernels such that the test runs with
# @all_solvers_kernels().
@pytest.mark.integtest  # run only as integration test
@all_solvers()
def test_infinite_time_independent_restart_pickle(solver_type, kernel_type):
    N = 10
    tmax = 5
    eps = 1.1  # 1E-8  # boundary valid up to time = tmax + 1 instead tmax
    salt = '1'
    E = 1.
    lat = kwant.lattice.square(norbs=2)
    syst = make_system_with_leads(lat, N, make_simple_lead)
    fsyst = syst.finalized()
    norbs = fsyst.site_ranges[-1][-1]

    boundaries = leads.automatic_boundary(fsyst.leads, tmax, params={'salt': salt})
    psi_st = kwant.wave_function(fsyst, energy=E, params={'time': 0, 'salt': salt})(0)[0]

    norbs = fsyst.site_ranges[-1][-1]
    psi_random = np.random.random((norbs,)) + 1j * np.random.random((norbs,))

    ### test restarting a calculation
    for psi_k, E in [(psi_random, None), (psi_st, E)]:
        psi = WaveFunction.from_kwant(fsyst, psi_k, boundaries=boundaries, params={'salt': salt}, energy=E,
                                      solver_type=solver_type, kernel_type=kernel_type)
        psi.evolve(tmax / 2)  # evolve forward a bit
        saved = pickle.dumps(psi)
        psi2 = pickle.loads(saved)
        psi.evolve(tmax - eps)
        psi2.evolve(tmax - eps)
        assert np.allclose(psi.psi(), psi2.psi())


@pytest.mark.integtest  # run only as integration test
@all_solvers_kernels()
def test_infinite_time_dependent(solver_type, kernel_type):
    N = 10
    salt = '1'
    lat = kwant.lattice.square(norbs=2)
    I0 = ta.identity(2)
    SZ = ta.array([[1, 0], [0, -1]])
    uniform = kwant.digest.uniform

    def td_onsite(site, time, salt):
        static_part = (4 + uniform(site.tag, salt=salt)) * I0
        td_part = SZ * sin(time)
        return static_part + td_part

    def phi_t(time, salt):
        return (1 - cos(time)) * ta.array([1, -1])  # diagonal part

    ### test when adding time-dependent voltage everywhere
    syst = make_td_system_with_leads(lat, N, make_simple_lead, td_onsite)
    # add TD voltage to leads as well
    leads.add_voltage(syst, 0, phi_t)
    leads.add_voltage(syst, 1, phi_t)
    # finalize
    fsyst = syst.finalized()
    nsites, _, norbs = fsyst.site_ranges[-1]

    sz_op = kwant.operator.Density(fsyst, SZ)
    boundaries = leads.automatic_boundary(fsyst.leads, tmax=11, params={'salt': salt})
    E = 1.

    # same time-dependent voltage everywhere, there should
    # be no change in the observables
    scattering_states = kwant.wave_function(fsyst, energy=E, params={'time': 0, 'salt': salt})
    psi_st = scattering_states(0)[0]
    perturbation_type = kernels.PerturbationExtractor
    psi = WaveFunction.from_kwant(syst=fsyst, boundaries=boundaries, params={'salt': salt},
                                  psi_init=psi_st, energy=E,
                                  perturbation_type=perturbation_type,
                                  solver_type=solver_type, kernel_type=kernel_type)
    should_be = sz_op(psi.psi())
    for time in (1., 5., 10.):
        psi.evolve(time)
        density = sz_op(psi.psi())
        assert np.allclose(
            should_be,
            density
        )


TIMESTEPS = [0, 1., 5]


def one_level_system(only_open=True):

    def make_system_2(a=1, gamma=1.0, W=1, L=30):

        lat = kwant.lattice.square(a=a, norbs=1)
        syst = kwant.Builder()

        def onsite_potential(site, time):
            return 4 * gamma * cmath.exp(- 1j * time)

        def gaussian(time, t0, A, sigma):
            return A * (1 + erf((time - t0) / sigma))

        # time dependent coupling with gaussian pulse
        def coupling_nn(site1, site2, time, t0, A, sigma):
            return - gamma * cmath.exp(- 1j * gaussian(time, t0, A, sigma))

        # Define the scattering region.
        syst[(lat(x, y) for x in range(L) for y in range(W))] = onsite_potential
        syst[lat.neighbors()] = -gamma
        # time dependent coupling between two sites 0 and 1
        for y in range(W):
            syst[lat(0, y), lat(1, y)] = coupling_nn

        # Define and attach the leads.
        # Construct the left lead.
        lead = kwant.Builder(kwant.TranslationalSymmetry((-a, 0)))
        lead[(lat(0, j) for j in range(W))] = 4 * gamma
        lead[lat.neighbors()] = -gamma

        # Attach the left lead and its reversed copy.
        syst.attach_lead(lead)
        syst.attach_lead(lead.reversed())

        return syst

    syst = make_system_2().finalized()
    if only_open:  # return only open mode
        return pytest.mark.parametrize('syst, energy, nb_open_modes',
                                       [(syst, 3, 1)])
    # return open and closed modes
    return pytest.mark.parametrize('syst, energy, nb_open_modes',
                                   [(syst, 3, 1), (syst, 0, 0)])


def closed_sytem():
    def make_system(a=1, t=1.0, r=10, r_time_dep=3):
        # Start with an empty tight-binding system and a single square lattice.
        # `a` is the lattice constant (by default set to 1 for simplicity).

        lat = kwant.lattice.square(a, norbs=1)

        syst = kwant.Builder()

        # Define the quantum dot
        def circle(r):
            def _(pos):
                (x, y) = pos
                rsq = x ** 2 + y ** 2
                return rsq < r ** 2
            return _

        # Formally takes the `time` parameter, but does not use it.
        def hopx(site1, site2, time, V, B):
            # The magnetic field is controlled by the parameter B
            # and oscillates in time
            y = site1.pos[1]
            if time > 0:
                return -t * np.exp(-1j * B * y)
            return 0

        # Decorated to notify tkwant that the function uses the `time` parameter.
        def potential(site, time, V, B):
            x, y = site.pos
            return 4 * t + np.sqrt(x**2 + y**2) * V * (1 - np.cos(time))

        syst[lat.shape(circle(r), (0, 0))] = 4 * t
        syst[lat.shape(circle(r_time_dep), (0, 0))] = potential
        # hoppings in x-direction
        syst[kwant.builder.HoppingKind((1, 0), lat, lat)] = hopx
        # hoppings in y-directions
        syst[kwant.builder.HoppingKind((0, 1), lat, lat)] = -t

        # It's a closed system, so no leads
        return syst

    syst = make_system().finalized()

    params = {'V': 0, 'B': 0}
    tparams = params.copy()
    tparams['time'] = 0

    H = syst.hamiltonian_submatrix(params=tparams)
    evals, evecs = np.linalg.eigh(H)

    energy = evals[0]
    psi_init = evecs[:, 0]

    psi_random = np.random.random(psi_init.shape) + 1j * np.random.random(psi_init.shape)

    return pytest.mark.parametrize('psi_init, syst, energy, params',
                                   [(psi_init, syst, energy, params),
                                    (psi_random, syst, None, params)])


def open_sytem():

    lat = kwant.lattice.square(norbs=1)

    def make_circle_system(radius):
        syst = kwant.Builder()
        syst[lat.shape(lambda pos: np.linalg.norm(pos) < radius, (0, 0))] = 4
        syst[lat.neighbors()] = -1
        return syst

    def make_lead(width):
        lead = kwant.Builder(kwant.TranslationalSymmetry((-1, 0)))
        lead[(lat(0, j) for j in range(-(width - 1) // 2, width // 2 + 1))] = 4
        lead[lat.neighbors()] = -1
        return lead

    syst = make_circle_system(radius=10)
    lead = make_lead(width=7)
    syst.attach_lead(lead)
    syst.attach_lead(lead.reversed())

    def faraday_flux(time, a, b):
        return a * (time - b * np.sin(0.1 * time)) / 2

    leads.add_voltage(syst, 0, faraday_flux)
    syst = syst.finalized()

    # create a time-dependent wavefunction that starts in a scattering state
    # originating from the left lead

    energy = 1.
    params = {'a': 0.1, 'b': 10}
    tparams = params.copy()
    tparams['time'] = 0
    scattering_states = kwant.wave_function(syst, energy=energy, params=tparams)
    psi_st = scattering_states(0)[0]

    psi_random = np.random.random(psi_st.shape) + 1j * np.random.random(psi_st.shape)

    return pytest.mark.parametrize('psi_init, syst, energy, params',
                                   [(psi_st, syst, energy, params),
                                    (psi_random, syst, None, params)])


@closed_sytem()
def test_WaveFunction_closed_system_params_copied(psi_init, syst, energy, params):

    params_copy = params.copy()
    wave_function = WaveFunction.from_kwant(psi_init=psi_init, syst=syst,
                                            energy=energy, params=params_copy)

    assert wave_function.params == params_copy
    wave_function.params = None
    assert params_copy == params


@open_sytem()
def test_WaveFunction_open_system_params_copied(psi_init, syst, energy, params):

    params_copy = params.copy()
    boundaries = leads.automatic_boundary(syst.leads, tmax=50, params=params)

    wave_function = WaveFunction.from_kwant(psi_init=psi_init, syst=syst, energy=energy,
                                            boundaries=boundaries, params=params_copy)

    assert wave_function.params == params_copy
    wave_function.params = None
    assert params_copy == params


@closed_sytem()
@pytest.mark.integtest
def test_WaveFunction_closed_system_default(psi_init, syst, energy, params):

    density_operator = kwant.operator.Density(syst)

    wave_function = WaveFunction.from_kwant(psi_init=psi_init, syst=syst,
                                            energy=energy, params=params)

    assert wave_function.time == 0
    assert wave_function.params == params

    if energy is None:
        assert wave_function.energy is None
    else:
        assert wave_function.energy == energy

    for time in TIMESTEPS:

        wave_function.evolve(time)
        assert wave_function.time == time
        assert wave_function.params == params

        density = wave_function.evaluate(density_operator)
        assert isinstance(density, np.ndarray)
        assert density.shape == psi_init.shape
        assert np.array([isinstance(x, float) for x in density]).all()

        assert isinstance(wave_function.psi(), np.ndarray)
        assert wave_function.psi().shape == psi_init.shape
        assert np.array([isinstance(x, complex) for x in wave_function.psi()]).all()


@open_sytem()
@pytest.mark.integtest
def test_WaveFunction_open_system_default(psi_init, syst, energy, params):

    density_operator = kwant.operator.Density(syst)
    tmax = 50
    boundaries = leads.automatic_boundary(syst.leads, tmax, params=params)

    wave_function = WaveFunction.from_kwant(psi_init=psi_init, syst=syst, energy=energy,
                                            boundaries=boundaries, params=params)

    assert wave_function.time == 0
    if energy is None:
        assert wave_function.energy is None
    else:
        assert wave_function.energy == energy

    for time in TIMESTEPS:

        assert time <= tmax
        wave_function.evolve(time)
        assert wave_function.time == time
        assert wave_function.params == params

        density = wave_function.evaluate(density_operator)
        assert isinstance(density, np.ndarray)
        assert density.shape == psi_init.shape
        assert np.array([isinstance(x, float) for x in density]).all()

        assert isinstance(wave_function.psi(), np.ndarray)
        assert wave_function.psi().shape == psi_init.shape
        assert np.array([isinstance(x, complex) for x in wave_function.psi()]).all()


@open_sytem()
def test_WaveFunction_open_system__init__raises(psi_init, syst, energy, params):

    # open system without boundaries
    pytest.raises(ValueError, WaveFunction.from_kwant, psi_init=psi_init, syst=syst,
                  energy=energy, params=params)

    boundaries = leads.automatic_boundary(syst.leads, tmax=6, params=params)

    params_with_time = params.copy()
    params_with_time['time'] = 0
    with pytest.raises(KeyError) as exc:
        WaveFunction.from_kwant(psi_init=psi_init, syst=syst, boundaries=boundaries,
                                energy=energy, params=params_with_time)
    error_string = str(exc.value).strip('\"')  # remove etra quotes
    assert error_string == "'params' must not contain time"

    # no params passed at all
    with pytest.raises(TypeError) as exc:
        WaveFunction.from_kwant(psi_init=psi_init, syst=syst, boundaries=boundaries,
                                energy=energy)
    assert str(exc.value) == 'System is missing required arguments: "a", "b"'

    params_not_matching = {'a': 0.1, 'B': 0.2}
    with pytest.raises(TypeError) as exc:
        WaveFunction.from_kwant(psi_init=psi_init, syst=syst, boundaries=boundaries,
                                energy=energy, params=params_not_matching)
    assert str(exc.value) == 'System is missing required arguments: "b"'


@closed_sytem()
def test_WaveFunction_closed_system__init__raises(psi_init, syst, energy, params):

    params_with_time = params.copy()
    params_with_time['time'] = 0
    with pytest.raises(KeyError) as exc:
        WaveFunction.from_kwant(psi_init=psi_init, syst=syst,
                                energy=energy, params=params_with_time)
    error_string = str(exc.value).strip('\"')  # remove etra quotes
    assert error_string == "'params' must not contain time"

    # no params passed at all
    with pytest.raises(TypeError) as exc:
        WaveFunction.from_kwant(psi_init=psi_init, syst=syst, energy=energy)
    assert str(exc.value) == 'System is missing required arguments: "V", "B"'

    params_not_matching = {'V': 0.1, 'b': 0.2}
    with pytest.raises(TypeError) as exc:
        WaveFunction.from_kwant(psi_init=psi_init, syst=syst,
                                energy=energy, params=params_not_matching)
    assert str(exc.value) == 'System is missing required arguments: "B"'


@open_sytem()
@pytest.mark.integtest
def test_WaveFunction_open_system_evolve_raises(psi_init, syst, energy, params):

    boundaries = leads.automatic_boundary(syst.leads, tmax=6, params=params)
    wave_function = WaveFunction.from_kwant(psi_init=psi_init, syst=syst, energy=energy,
                                            boundaries=boundaries, params=params)
    wave_function.evolve(time=5)

    # evolve back in time
    backward_time = wave_function.time - 0.5
    pytest.raises(ValueError, wave_function.evolve, time=backward_time)


@closed_sytem()
@pytest.mark.integtest
def test_WaveFunction_closed_system_evolve_raises(psi_init, syst, energy, params):

    wave_function = WaveFunction.from_kwant(psi_init=psi_init, syst=syst,
                                            energy=energy, params=params)

    wave_function.evolve(time=5)
    # evolve back in time
    backward_time = wave_function.time - 0.5
    pytest.raises(ValueError, wave_function.evolve, time=backward_time)


@open_sytem()
def test_WaveFunction_open_system_evaluate_raises(psi_init, syst, energy, params):

    boundaries = leads.automatic_boundary(syst.leads, tmax=6, params=params)
    wave_function = WaveFunction.from_kwant(psi_init=psi_init, syst=syst,
                                            energy=energy, boundaries=boundaries, params=params)

    # bind should fail
    def radius(site):
        x, y = site.pos
        return np.sqrt(x**2 + y**2)
    density_operator = kwant.operator.Density(syst, radius).bind()

    with pytest.raises(ValueError) as exc:
        wave_function.evaluate(density_operator)
    assert str(exc.value) == 'Operator must not use pre-bind values'


@closed_sytem()
def test_WaveFunction_closed_system_evaluate_raises(psi_init, syst, energy, params):

    wave_function = WaveFunction.from_kwant(psi_init=psi_init, syst=syst, energy=energy,
                                            params=params)

    # bind should fail
    def radius(site):
        x, y = site.pos
        return np.sqrt(x**2 + y**2)
    density_operator = kwant.operator.Density(syst, radius).bind()

    with pytest.raises(ValueError) as exc:
        wave_function.evaluate(density_operator)
    assert str(exc.value) == 'Operator must not use pre-bind values'


@one_level_system(only_open=False)
def test_scattering_states_default(syst, energy, nb_open_modes):

    tmax = 50
    params = {'t0': 0, 'A': 0.00157, 'sigma': 24}
    for lead in range(len(syst.leads)):
        states = ScatteringStates(syst, energy, lead, tmax, params=params)
        assert states.energy == energy
        assert states.lead == lead
        assert isinstance(states, collections.abc.Iterable)
        assert len(states) == nb_open_modes
        assert np.array([isinstance(state, WaveFunction) for state in states]).all()

        # check added attribute in each instance
        for mode, state in enumerate(states):
            assert isinstance(state.lead, int)
            assert isinstance(state.mode, int)
            assert state.lead == lead
            assert state.mode == mode


@one_level_system()
@pytest.mark.parametrize('time', TIMESTEPS)
def test_scattering_states_against_reference(syst, energy, nb_open_modes, time):

    # test only for one lead/mode pair, as calculations are expensive
    lead = 0
    mode = 0
    params = {'t0': 0, 'A': 0.00157, 'sigma': 24}

    tmax = 50
    boundaries = leads.automatic_boundary(syst.leads, tmax, params=params)
    assert time <= tmax

    states = ScatteringStates(syst, energy, lead, tmax, params=params)
    state = states[mode]
    state.evolve(time)

    # reference result
    tparams = params.copy()
    tparams['time'] = 0
    scattering_states = kwant.wave_function(syst, energy=energy, params=tparams)
    psi_st = scattering_states(lead)[mode]
    psi_ref = WaveFunction.from_kwant(psi_init=psi_st, syst=syst, boundaries=boundaries, energy=energy, params=params)
    psi_ref.evolve(time)

    assert np.allclose(state.psi(), psi_ref.psi())


@one_level_system()
def test_scattering_states__init__raises(syst, energy, nb_open_modes):

    lead = 0
    params = {'t0': 0, 'A': 0.00157, 'sigma': 24}
    tmax = 50
    boundaries = leads.automatic_boundary(syst.leads, tmax, params=params)

    # system has no leads
    system_without_leads = copy.deepcopy(syst)
    delattr(system_without_leads, 'leads')
    pytest.raises(AttributeError, ScatteringStates, syst=system_without_leads,
                  energy=energy, lead=lead, tmax=tmax, params=params)
    # energy is not a float
    pytest.raises(TypeError, ScatteringStates, syst=syst,
                  energy=energy + 1j, lead=lead, tmax=tmax, params=params)
    # lead index is not an integer
    pytest.raises(TypeError, ScatteringStates, syst=syst,
                  energy=energy, lead=float(lead), tmax=tmax, params=params)
    # lead index is out of range
    pytest.raises(ValueError, ScatteringStates, syst=syst,
                  energy=energy, lead=len(syst.leads) + 1, tmax=tmax, params=params)
    # tmax/boundaries are both present, but mutually exclusive
    pytest.raises(ValueError, ScatteringStates, syst=syst,
                  energy=energy, lead=lead, tmax=tmax, boundaries=boundaries, params=params)

    params_with_time = {'t0': 0, 'A': 0.00157, 'sigma': 24, 'time': 0}
    with pytest.raises(KeyError) as exc:
        ScatteringStates(syst, lead=0, energy=2.4, tmax=tmax,
                         params=params_with_time)
    error_string = str(exc.value).strip('\"')  # remove etra quotes
    assert error_string == "'params' must not contain time"

    # no params passed at all
    with pytest.raises(TypeError) as exc:
        ScatteringStates(syst, lead=0, energy=2.4, tmax=tmax)
    assert str(exc.value) == 'System is missing required arguments: "t0", "A", "sigma"'

    params_without_sigma = {'t0': 0, 'A': 0.00157}
    with pytest.raises(TypeError) as exc:
        ScatteringStates(syst, lead=0, energy=2.4, tmax=tmax,
                         params=params_without_sigma)
    assert str(exc.value) == 'System is missing required arguments: "sigma"'


@one_level_system()
def test_scattering_states__getitem__raises(syst, energy, nb_open_modes):
    params = {'t0': 0, 'A': 0.00157, 'sigma': 24}
    states = ScatteringStates(syst, energy, lead=0, tmax=50, params=params)
    # mode index out of range
    pytest.raises(KeyError, lambda: states[len(states) + 1])
    # mode index is not an integer
    pytest.raises(TypeError, lambda: states[float(len(states))])


@pytest.mark.integtest
@all_solvers_kernels()
def test_time_argument_name_and_initial_time_change(solver_type, kernel_type):

    N = 10
    lat = kwant.lattice.square(norbs=2)
    I0 = ta.identity(2)
    SX = ta.array([[0, 1], [1, 0]])
    uniform = kwant.digest.uniform

    # make twice a system, one where time time variable is named time
    # and the initial time starts at time=0 (the default case)
    # and a second system where the time varible is named zeit
    # and the initial time starts at zeit = 42

    def td_onsite_time(site, time, salt):
        static_part = (4 + uniform(site.tag, salt=salt)) * I0
        td_part = uniform(site.tag, salt=salt + '1') * SX * cos(time)
        return static_part + td_part

    zeit_offset = 42  # the initial time

    def td_onsite_zeit(site, zeit, salt):
        time = zeit - zeit_offset
        static_part = (4 + uniform(site.tag, salt=salt)) * I0
        td_part = uniform(site.tag, salt=salt + '1') * SX * cos(time)
        return static_part + td_part

    fsyst_time = make_td_system(lat, N, td_onsite_time).finalized()
    fsyst_zeit = make_td_system(lat, N, td_onsite_zeit).finalized()

    # the time is named time and starts at zero
    H0_time = fsyst_time.hamiltonian_submatrix(params={'time': 0, 'salt': '1'})
    # the scipy.sparse.linalg.eigsh routine is not re-entrant
    evals_time, evecs_time = np.linalg.eigh(H0_time)
    psi_st_time = evecs_time[:, 0]
    E_time = evals_time[0]

    psi_time = WaveFunction.from_kwant(syst=fsyst_time, params={'salt': '1'},
                                       psi_init=psi_st_time, energy=E_time,
                                       solver_type=solver_type, kernel_type=kernel_type)

    assert psi_time.time == 0
    assert psi_time.time_start == 0
    assert psi_time.time_name == 'time'

    # the time is named zeit and starts at 42
    H0_zeit = fsyst_zeit.hamiltonian_submatrix(params={'zeit': zeit_offset, 'salt': '1'})
    evals_zeit, evecs_zeit = np.linalg.eigh(H0_zeit)
    psi_st_zeit = evecs_zeit[:, 0]
    E_zeit = evals_zeit[0]

    psi_zeit = WaveFunction.from_kwant(syst=fsyst_zeit, params={'salt': '1'},
                                       psi_init=psi_st_zeit, energy=E_zeit,
                                       solver_type=solver_type, kernel_type=kernel_type,
                                       time_name='zeit', time_start=zeit_offset)

    assert psi_zeit.time == zeit_offset
    assert psi_zeit.time_start == zeit_offset
    assert psi_zeit.time_name == 'zeit'

    # check that initial condition is similar
    assert np.allclose(psi_st_time, psi_st_zeit)
    assert E_time == E_zeit

    for time in (0., 1., 5.):
        psi_time.evolve(time)
        psi_zeit.evolve(time + zeit_offset)

        assert psi_time.time_start == 0
        assert psi_zeit.time_start == zeit_offset
        assert psi_time.time == time
        assert psi_zeit.time == time + zeit_offset

        assert np.allclose(psi_time.psi(), psi_zeit.psi())

    for t0 in [1 + 1j, 'a', None, -float('inf'), float('inf')]:
        with pytest.raises(TypeError) as exc:
            WaveFunction.from_kwant(syst=fsyst_time, params={'salt': '1'},
                                    psi_init=psi_st_time, energy=E_time,
                                    solver_type=solver_type, kernel_type=kernel_type,
                                    time_start=t0)
        assert str(exc.value) == "time must be a finite real number"


@pytest.mark.integtest
def test_time_name_and_start_default_change_scattering_state():
    N = 10
    lat = kwant.lattice.square(norbs=2)
    I0 = ta.identity(2)
    SZ = ta.array([[1, 0], [0, -1]])
    uniform = kwant.digest.uniform

    # test only for one lead/mode pair, as calculations are expensive
    tmax = 100
    lead = 0
    mode = 0
    energy = 1
    params = {'salt': '1'}

    def td_onsite_time(site, time, salt):
        static_part = (4 + uniform(site.tag, salt=salt)) * I0
        td_part = SZ * sin(time)
        return static_part + td_part

    def phi_time(time, salt):
        return (1 - cos(time)) * ta.array([1, -1])  # diagonal part

    zeit_offset = 42  # the initial time

    def td_onsite_zeit(site, zeit, salt):
        time = zeit - zeit_offset
        static_part = (4 + uniform(site.tag, salt=salt)) * I0
        td_part = SZ * sin(time)
        return static_part + td_part

    def phi_zeit(zeit, salt):
        time = zeit - zeit_offset
        return (1 - cos(time)) * ta.array([1, -1])  # diagonal part

    ### test when adding time-dependent voltage everywhere
    syst = make_td_system_with_leads(lat, N, make_simple_lead, td_onsite_time)
    leads.add_voltage(syst, 0, phi_time)
    leads.add_voltage(syst, 1, phi_time)
    fsyst_time = syst.finalized()

    syst = make_td_system_with_leads(lat, N, make_simple_lead, td_onsite_zeit)
    leads.add_voltage(syst, 0, phi_zeit)
    leads.add_voltage(syst, 1, phi_zeit)
    fsyst_zeit = syst.finalized()

    # default
    states_time = ScatteringStates(fsyst_time, energy, lead, tmax, params=params)
    state_time = states_time[mode]

    assert state_time.time == 0
    assert state_time.time_start == 0
    assert state_time.time_name == 'time'

    # changed
    WaveFunction_zeit = ft.partial(WaveFunction.from_kwant, time_name='zeit',
                                   time_start=zeit_offset)

    states_zeit = ScatteringStates(fsyst_zeit, energy, lead, tmax, params=params,
                                   wavefunction_type=WaveFunction_zeit)
    state_zeit = states_zeit[mode]

    assert state_zeit.time == zeit_offset
    assert state_zeit.time_start == zeit_offset
    assert state_zeit.time_name == 'zeit'

    for time in (0., 1., 5.):
        state_time.evolve(time)
        state_zeit.evolve(time + zeit_offset)

        assert state_time.time_start == 0
        assert state_zeit.time_start == zeit_offset
        assert state_time.time == time
        assert state_zeit.time == time + zeit_offset

        assert np.allclose(state_time.psi(), state_zeit.psi())

    for t0 in [1 + 1j, 'a', None, -float('inf'), float('inf')]:
        WaveFunction_zeit = ft.partial(WaveFunction.from_kwant, time_name='zeit', time_start=t0)
        with pytest.raises(TypeError) as exc:
            ScatteringStates(fsyst_zeit, energy, lead, tmax, params=params,
                             wavefunction_type=WaveFunction_zeit)
        assert str(exc.value) == "time must be a finite real number"


def test_wavefunction_from_hamiltonian():
    def create_system(L):

        # system building
        lat = kwant.lattice.square(a=1, norbs=1)
        syst = kwant.Builder()

        # central scattering region
        syst[(lat(x, 0) for x in range(L))] = 1
        syst[lat.neighbors()] = -1

        return syst

    num_sites = 400
    syst = create_system(num_sites).finalized()

    # lattice sites and time steps
    xi = np.arange(num_sites)
    times = np.arange(0, 600, 50)

    # initial condition
    k = np.pi / 6
    psi_init = np.exp(- 0.001 * (xi - 100)**2 + 1j * k * xi)

    # hamiltonian matrix
    diag = np.ones(len(xi))
    offdiag = - np.ones(len(xi) - 1)
    H0 = scipy.sparse.diags([diag, offdiag, offdiag],
                            [0, 1, -1], dtype=complex)

    H0_kwant = syst.hamiltonian_submatrix(params={'time': 0}, sparse=True)

    assert_array_almost_equal(H0.todense(), H0_kwant.todense())

    wave_func = WaveFunction(H0, W=None, psi_init=psi_init)

    wave_func_kwant = WaveFunction.from_kwant(syst, psi_init)

    for time in times:
        wave_func.evolve(time)
        wave_func_kwant.evolve(time)
        psi = wave_func.psi()
        psi_kwant = wave_func_kwant.psi()
        assert_array_almost_equal(psi, psi_kwant)


def test_wavefunction_from_hamiltonian_raises():
    num_sites = 400

    # lattice sites and time steps
    xi = np.arange(num_sites)

    # initial condition
    k = np.pi / 6
    psi_init = np.exp(- 0.001 * (xi - 100)**2 + 1j * k * xi)

    # hamiltonian matrix which is too small
    diag = np.ones(len(xi) - 1)
    offdiag = - np.ones(len(xi) - 2)
    H0 = scipy.sparse.diags([diag, offdiag, offdiag], [0, 1, -1])

    with pytest.raises(ValueError) as exc:
        WaveFunction(H0, W=None, psi_init=psi_init)
    assert 'initial condition size=400 is larger than the Hamiltonian matrix H0 size=399' in str(exc.value)

    # hamiltonian matrix
    diag = np.ones(len(xi) - 1)
    offdiag = - np.ones(len(xi) - 2)
    H0 = scipy.sparse.diags([diag, offdiag, offdiag], [0, 1, -1])

    # perturbation which is too small
    diag = np.ones(len(xi))
    offdiag = - np.ones(len(xi) - 1)
    H0 = scipy.sparse.diags([diag, offdiag, offdiag], [0, 1, -1])

    def W(*args, **kwargs):
        pass
    W.size = num_sites - 1

    with pytest.raises(ValueError) as exc:
        WaveFunction(H0, W, psi_init=psi_init)
    assert 'initial condition size=400 must be equal to the perturbation W size=399' in str(exc.value)



# -*- coding: utf-8 -*-
# Copyright 2020 - 2023 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/license.html.
# A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/authors.html.
"""Tools for calculating interacting systems."""

import numpy as np
import scipy

import kwant

from . import onebody, mpi, _logging, _common

_all__ = ['AdaptiveStepsize', 'Extrapolate', 'SelfConsistentState']


# set module logger
logger = _logging.make_logger(name=__name__)
log_func = _logging.log_func(logger)


@log_func
class AdaptiveStepsize:
    def __init__(self, atol, rtol, tau_min=1E-6, fac=0.9, facmin=0.1, facmax=3, p=1):
        """A class to estimate the stepsize tau."""

        assert atol >= 0
        assert rtol >= 0
        assert tau_min > 0

        self.atol = atol
        self.rtol = rtol
        self.tau_min = tau_min
        self.fac = fac
        self.facmin = facmin
        self.facmax = facmax
        self.a = - 1 / (1 + p)  # order p

    def __call__(self, y, dyt, tau):
        # Eq. (4.13), page 168 in book:
        # E. Hairer, S. Nørsett, and G. Wanner, Solving Ordinary Differential Equations,
        # I: Nonstiff Problems, Second  revised edition (Springer, Berlin, Heidelberg, 2008).
        assert tau > 0
        error = self._estimate_error(y, dyt)

        if error > 0:
            new_tau = tau * min(self.facmax, max(self.facmin, self.fac * error**self.a))
        else:  # error is zero
            new_tau = tau * min(self.facmax, self.facmin)

        if new_tau < self.tau_min:
            logger.warning('minimal self-consistent stepsize tau_min={} reached.'
                           .format(self.tau_min))
            new_tau = self.tau_min
        return new_tau

    def _estimate_error(self, y, dyt):
        scale = self.atol + self.rtol * np.abs(y)
        return np.max(np.abs(dyt) / scale)


@log_func
class Extrapolate:
    r"""Extrapolate a function y = f(x) in positive x direction.

       To construct an approximation g(x), the function f(x) must be sampled
       on different points (x0 < x1 < x2, < ...). Sampling points are not required
       to be equidistant. For an accurate approximation g(x) \approx f(x),
       the function f(x) must be smooth and vary little between sampling points.
       In addition, the distance between the evaluation point x and
       the last sampling point x_n must fulfill dx = x - x_n << 1.
       Extrapolation is based on polynominal approximation g(x).

       g(x) = \sum_{n=0}^m a_n x

    """
    def __init__(self, y0, tau, order=1, x0=0, dy0=0, tol=1E-14):
        """Initialize an extrapolation function g(x)"""

        y0 = np.asarray(y0)
        dy = np.asarray(dy0)

        if not _common.is_type(x0, 'real_number'):
            raise TypeError('argument x0 must be a real scalar value')
        if not _common.is_type(tau, 'real_number') or tau <= 0:
            raise ValueError('update timestep must be real and larger zero, found tau={}'.format(tau))
        if not _common.is_type_array(y0, 'number').all():
            raise TypeError('y0 must be a scalar or vector of numbers')
        if not _common.is_type_array(dy, 'number').all():
            raise TypeError('dy must be a scalar or vector of numbers')
        if not _common.is_type(tol, 'real_number') or tol < 0:
            raise ValueError('tol must real and positive, found tol={}'.format(tol))

        self.x0 = x0
        self.x1 = x0 + tau
        self.order = order

        self._y0 = y0
        self._yt0 = y0
        self._dy = dy
        self._tol = tol

    def _g_func(self, x):
        if not _common.is_type_array(x, 'real_number').all():
            raise TypeError('argument must be a scalar or vector of real numbers')
        if not ((x >= self.x0 - self._tol) & (x <= self.x1 + self._tol)):
            raise ValueError('argument x={} is not inside interval=[{}, {}]'
                             .format(x, self.x0, self.x1))
        if self.order == 0:
            return self._y0
        elif self.order == 1:
            return self._yt0 + (x - self.x0) * self._dy
        else:
            raise NotImplementedError("extrapolation order={} is not "
                                      "implemented".format(self.order))

    def __call__(self, x):
        """Return the extrapolation function g(x)"""
        x = np.asarray(x)
        if x.size == 1:
            return self._g_func(x)
        return np.array([self._g_func(xx) for xx in x])

    def add_point(self, x, y):
        """Add a new point on the rhs"""
        y = np.asarray(y)
        if not _common.is_type(x, 'real_number'):
            raise TypeError('argument x must be a real scalar value')
        if not _common.is_type_array(y, 'number').all():
            raise TypeError('y must be a scalar or vector of numbers')
        if not _common.is_zero(x - self.x1, self._tol):
            raise ValueError('update time must be at x={}, found x={}'.format(self.x1, x))

        yt = self.__call__(x)

        dx = x - self.x0
        dy = y - self._y0
        dyt = y - yt

        # update the stored evaluation points
        self.x0 = x
        self._y0 = y
        self._yt0 = yt
        self._dx = dx
        self._dy = dy
        self._dyt = dyt

        return dyt

    def set_stepsize(self, tau):

        if not _common.is_type(tau, 'real_number') or tau <= 0:
            raise ValueError('update timestep must be real and larger zero, found tau={}'.format(tau))

        if self.order == 1:
            self._dy = self._dy / self._dx + self._dyt / tau

        # update the stored evaluation points
        self.x1 = self.x0 + tau
        return


@log_func
class SelfConsistentState:
    """Solve self-consistently the interacting and time-dependent many-particle
       Schrödinger equation"

       An instance of the class behaves like a manybody wavefunction.
       The class is independent from the physical decoupling of the interaction
       and manages only the self-consistent update of the memory kernel.
       """
    def __init__(self, wave_function, mf_operator, mf_potential,
                 atol=1E-6, rtol=1E-6, tau=AdaptiveStepsize,
                 extrapolator_type=Extrapolate):
        r"""
        Parameters
        ----------
        wave_function : `tkwant.manybody.WaveFunction`
            The manybody wavefunction of the non-interacting system.
        mf_operator : callable
            An operator to evaluate the expectation values needed for to
            calculate the mean-field potential.
            Must have the calling signature of `kwant.operator`
            and return a numpy array.
        mf_potential : object
            An object to calculate the mean-field potential
            Must have a ``prepare()`` and an ``evaluate()`` method.
        atol : float, optional
            Absolute tolerance for the ``mf_operator`` function extrapolation.
        rtol : float, optional
            Relative tolerance for the `mf_operator`` function extrapolation.
        tau : float or ``tkwant.interaction.AdaptiveStepsize``, optional
            For adaptive stepsize control, tau is a function to estimate the
            extrapolation stepsize. If ``tau`` is set to
            a number, the value is used as a constant stepsize.
        extrapolator_type : ``tkwant.interaction.Extrapolate``, optional
            A class to extrapolate the self-consistent potential
        """
        self.wavefunction = wave_function
        self._mf_operator = mf_operator
        self._mf_potential = mf_potential
        self.steps = 0

        self.time = wave_function.time  # time of the manybody state

        # tau is the stepsize from last self-consistent update (at time_sc - tau)
        # time_sc is the time of the next self-consistent update,
        # we set it initially to the minimal tau from the stepsize estimator
        if _common.is_type(tau, 'real_number'):  # fixed stepsize tau
            if tau <= 0:
                raise ValueError("stepsize tau must be > 0")
            self._tau = tau

            def const_tau(*args, **kwargs):
                return tau
            self._estimate_tau = const_tau
        else:  # adaptive stepsize tau
            self._estimate_tau = tau(rtol, atol)
            self._tau = self._estimate_tau.tau_min
        self._time_sc = self.time + self._tau

        # evaluate the operator and construct an extrapolation function f(t)
        # calculate the actual mean field potential Q(t) from f(t)
        # add Q(t) to the tkwant hamiltonian
        try:
            y = self.wavefunction.evaluate(self._mf_operator, root=None)
        except TypeError:  # if root argument not present
            y = self.wavefunction.evaluate(self._mf_operator)
        self._yt = extrapolator_type(y, self._tau, x0=self.time)
        self._mf_potential.prepare(self._yt, self._time_sc)
        self.wavefunction.add_perturbation(self._mf_potential.evaluate)

    def evolve(self, time):
        """
        Evolve all the self-consistent manybody state up to ``time``.

        Parameters
        ----------
        time : int or float
            time argument up to which the state should be evolved

        Notes
        -----
        The self-consistent updates are completely independent from the
        requested timesteps ``time``.
        The reported error max|y - yt| is between the operator value y and
        its extrapolation yt, estimated at tmax of the extrapolation interval.
        """
        # perform self-consistent updates in foreward-time direction
        # until the next self-consistent update time ``time_sc`` lies
        # ahead of the actual requested time ``time`` for the manybody state
        while self._time_sc <= time:

            self.wavefunction.evolve(self._time_sc)

            # calculate the exact mean-field potential at ``time_sc``,
            # and estimate the error of the extrapolation function
            try:
                y = self.wavefunction.evaluate(self._mf_operator, root=None)
            except TypeError:  # if root argument not present
                y = self.wavefunction.evaluate(self._mf_operator)
            dyt = self._yt.add_point(self._time_sc, y)

            # estimate the stepsize tau and the time for the next self-consistent update
            dt = self._estimate_tau(y, dyt, self._tau)
            logger.debug('tmin, tmax= [{}, {}], new tau= {}, max|y - yt|= {}'
                         .format(self._time_sc - self._tau, self._time_sc, dt, np.max(np.abs(dyt))))
            self._tau = dt
            time_next = self._time_sc + self._tau

            # update the self-consistent mean-field extrapolation
            # and update the mean-field potential in the manybody wave-function
            # by providing the new extrapolation function
            self._yt.set_stepsize(self._tau)
            self._mf_potential.prepare(self._yt, time_next)
            self.wavefunction.add_perturbation(self._mf_potential.evaluate)
            self._time_sc = time_next

            self.steps += 1

        self.wavefunction.evolve(time)
        self.time = time

    def evaluate(self, operator, root=0):
        """Evaluate the expectation value of an operator at the current time.

        Arguments and return values similar to `tkwant.manybody.WaveFunction`
        """
        try:
            return self.wavefunction.evaluate(operator, root)
        except TypeError:  # onebody wavefunction has no root argument
            return self.wavefunction.evaluate(operator)



# Script 1/2 to produce the figure in Tkwant's "DC current through a 1D chain" tutorial.
#
# Step 1: run the simulation (chem_vs_elec_bias_run_computation.py)
# Step 2: plot the data (chem_vs_elec_bias_plot_results.py)
#
# This is the script for step 1 to do the kwant/tkwant simulation.
# Run this script preferably on a cluster using MPI parallelization,
# because it will take some time.
# The result from the calculation will be saved to a file.

import tkwant
import kwant

import cmath
import scipy
import numpy as np
import pickle
import sys
import functools


# --- MPI helper routines

def am_master():
    """Return true for the MPI master rank"""
    return tkwant.mpi.get_communicator().rank == 0


def print_master(*args, **kwargs):
    if am_master():
        print(*args, **kwargs)
    sys.stdout.flush()


# --- system definition for electric and chemical potential in steady state regime

def make_system_static(a=1):

    def onsite(site, ve):
        return ve

    # central system
    lat = kwant.lattice.square(a=a, norbs=1)
    syst = kwant.Builder()

    syst[(lat(i, 0) for i in [-1, 0, 1])] = 0
    syst[lat.neighbors()] = -1

    # leads
    lead_left = kwant.Builder(kwant.TranslationalSymmetry((-a, 0)))

    lead_left[lat(0, 0)] = onsite
    lead_left[lat.neighbors()] = -1
    syst.attach_lead(lead_left)

    lead_right = kwant.Builder(kwant.TranslationalSymmetry((-a, 0)))

    lead_right[lat(0, 0)] = 0
    lead_right[lat.neighbors()] = -1

    syst.attach_lead(lead_right.reversed())

    return syst, lat


# --- system definition for electric and chemical potential in transient regime

def coupling_nn(site1, site2, time, ve):
    phi = ve * time
    return - cmath.exp(- 1j * phi)


def make_system_transient(a=1):

    lat = kwant.lattice.square(a=a, norbs=1)
    syst = kwant.Builder()

    syst[(lat(i, 0) for i in [-1, 0, 1])] = 0
    syst[lat.neighbors()] = -1

    # time dependent coupling - e^(-i \phi(t)) c^\dagger_0 c_{-1}
    syst[(lat(0, 0), lat(-1, 0))] = coupling_nn

    lead = kwant.Builder(kwant.TranslationalSymmetry((-a, 0)))
    lead[lat(0, 0)] = 0
    lead[lat.neighbors()] = -1
    syst.attach_lead(lead)
    syst.attach_lead(lead.reversed())

    return syst, lat


# --- current from scattering matrix for electrical and chemical potential

def current_vc(vc):
    return min(2, vc)


def transmission(syst, energy, ve):
    try:
        smatrix = kwant.smatrix(syst, energy=energy, params={"ve": ve})
        res = smatrix.transmission(1, 0)
    except:  # exactly at the band opening, the kwant smatrix crashes
        res = 0
    return res


def current_ve(syst, ve):
    def transfunc(energy):
        return transmission(syst, energy, ve)
    return scipy.integrate.quad(transfunc, 0, ve)[0]


def main():

    # define the system to calculate the steady-state current
    syst_stat, lat = make_system_static()
    syst_stat = syst_stat.finalized()

    # define the system to calculate the transient current
    syst_trans, lat = make_system_transient()
    syst_trans = syst_trans.finalized()

    # parameters
    mu = 0
    tmax = 100

    all_results = {}

    # current from site 0 to site 1
    current_operator = kwant.operator.Current(syst_stat, where=[(lat(1, 0), lat(0, 0))])

    # loop over chemical/electrical potential, vc = ve = v
    potential = np.append(np.logspace(-6, -1, 40), np.linspace(0.1, 5, 50))
    for v in potential:

        print_master('v= {}'.format(v))
        results = {}

        # occupation of left and right lead
        occupations = [tkwant.manybody.lead_occupation(chemical_potential=mu+v),
                       tkwant.manybody.lead_occupation(chemical_potential=mu)]

        # --- (Ia) - chemical potential - scattering matrix
        results['I_vc_sc'] = current_vc(v)

        # --- (Ib) - chemical potential - wave function
        state = tkwant.manybody.State(syst_stat, tmax=1, occupations=occupations, params={"ve": 0})
        current = state.evaluate(current_operator)
        if am_master():
            results['I_vc_wf'] = 2 * np.pi * current

        # --- (IIa) - electrical potential - scattering matrix
        results['I_ve_sc'] = current_ve(syst_stat, v)

        # --- (IIb) - electrical potential - wave function
        state = tkwant.manybody.State(syst_stat, tmax=1, occupations=occupations, params={"ve": v})
        current = state.evaluate(current_operator)
        if am_master():
            results['I_ve_wf'] = 2 * np.pi * current

        # --- (IIc) - electrical potential - wave function - transient regime

        # turn off extrapolation of the time-dependent perturbation
        # this becomes crutial for large values of v
        onebody_wavefunction_type = functools.partial(tkwant.onebody.WaveFunction.from_kwant,
                                                      perturbation_type=tkwant.onebody.kernels.PerturbationExtractor)
        scattering_state_type = functools.partial(tkwant.onebody.ScatteringStates,
                                                  wavefunction_type=onebody_wavefunction_type)

        # defining no occupation corresponds to mu=0 in all leads
        state = tkwant.manybody.State(syst_trans, tmax=tmax, params={'ve': v},
                                      scattering_state_type=scattering_state_type)
        state.evolve(tmax)
        state.refine_intervals()
        current = state.evaluate(current_operator)
        if am_master():
            results['I_ve_wf_tr'] = 2 * np.pi * current

        all_results[v] = results

    #  write the result into a file
    if am_master():
        pickle.dump(all_results, open('chem_vs_elec_bias_result_data.npy', "wb"))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright 2019 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/license.html.
# A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/authors.html.

"""Test module for `tkwant._common`"""

import functools
import pytest

from .._common import get_default_function_argument


def foo(a, b=42, c=None):
    pass


def bar(x, a, b=42, c=None):
    pass


class Foo():
    def __init__(self, a, b=42, c=None):
        pass


class Bar():
    def __init__(self, x, a, b=42, c=None):
        pass


FUNCTIONS = [foo, bar, Foo, Bar]
FUNCTION_NAMES = ['foo', 'bar', 'Foo', 'Bar']


@pytest.mark.parametrize('foo, function_name', zip(FUNCTIONS, FUNCTION_NAMES))
def test_get_default_function_argument_with_function(foo, function_name):

    assert get_default_function_argument(foo, 'b') == 42
    assert get_default_function_argument(foo, 'c') is None

    # check that partial binding works
    foo_bind = functools.partial(foo, a=2)

    assert get_default_function_argument(foo_bind, 'a') == 2
    assert get_default_function_argument(foo_bind, 'b') == 42
    assert get_default_function_argument(foo_bind, 'c') is None

    # check if argument has no default value
    with pytest.raises(ValueError) as exc:
        get_default_function_argument(foo, 'a')
    assert str(exc.value) == "Function argument a not present in {}"\
                             .format(function_name)

    # check if argument has not default value in the prebind version
    with pytest.raises(ValueError) as exc:
        get_default_function_argument(foo_bind, 'x')
    assert str(exc.value) == "Function argument x not present in <unknown func>"

    # check if argument is not present
    with pytest.raises(ValueError) as exc:
        get_default_function_argument(foo, 'd')
    assert str(exc.value) == "Function argument d not present in {}"\
                             .format(function_name)

    # check if argument is not present in the prebind version
    with pytest.raises(ValueError) as exc:
        get_default_function_argument(foo_bind, 'd')
    assert str(exc.value) == "Function argument d not present in <unknown func>"



import tkwant
import kwant
import cmath
from scipy.special import erf
import matplotlib.pyplot as plt


def am_master():
    """Return true for the MPI master rank"""
    return tkwant.mpi.get_communicator().rank == 0


def make_system(L=400):

    def gaussian(time, t0=40, A=1.57, sigma=24):
        return A * (1 + erf((time - t0) / sigma))

    # time dependent coupling with gaussian pulse
    def coupling_nn(site1, site2, time):
        return - cmath.exp(- 1j * gaussian(time))

    # system building
    lat = kwant.lattice.square(a=1, norbs=1)
    syst = kwant.Builder()

    # central scattering region
    syst[(lat(x, 0) for x in range(L))] = 1
    syst[lat.neighbors()] = -1
    # time dependent coupling between two sites in the center
    syst[lat(L // 2, 0), lat(L // 2 - 1, 0)] = coupling_nn

    # add leads
    sym = kwant.TranslationalSymmetry((-1, 0))
    lead_left = kwant.Builder(sym)
    lead_left[lat(0, 0)] = 1
    lead_left[lat.neighbors()] = -1
    syst.attach_lead(lead_left)
    syst.attach_lead(lead_left.reversed())

    return syst


def main():

    syst = make_system().finalized()
    sites = [site.pos[0] for site in syst.sites]
    times = [40, 80, 120, 160]

    density_operator = kwant.operator.Density(syst)

    state = tkwant.manybody.State(syst, max(times))

    density0 = state.evaluate(density_operator)

    for time in times:
        state.evolve(time=time)
        if time == 40:
            state.refine_intervals()
        error = state.estimate_error()
        density = state.evaluate(density_operator)
        if am_master():
            print('time={}, error={:10.4e}'.format(time, error))
            plt.plot(sites, density - density0, label='time={}'.format(time))

    if am_master():
        plt.legend()
        plt.xlabel(r'site position $i$')
        plt.ylabel(r'charge density $n$')
        plt.show()


if __name__ == '__main__':
    main()
